/**
  Lighting demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-01-31 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/mat4.h"
using osl::mat4;
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "ogl/image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "ogl/image.cpp"

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

#include "GL/glui.h" /* OpenGL User Interface library (for making a GUI) */

void oglExit(const char *what,const char *why) {
	printf(what,why);
	exit(1);
}
int nPotato=10;
float walking_speed=30.0;
int move_camera=1;
mat4 light_direction=mat4(1.0);

/******************* Object, read from .obj file ***************/
class obj_mesh {
public:
	obj_mesh(const char *fName);
	
	void draw(void);

private:
	std::vector<vec3> vtx; /* 3D location of vertices */
	std::vector<vec3> norm; /* surface normal (unit length) */
	
	struct myFace {
		int v[4]; /* 1-based vertex index */
		int n[4]; /* 1-based normal vector index */
	};
	void draw_face(const myFace &f,int nVertices);
	std::vector<myFace> quad;
	std::vector<myFace> tri;
};

obj_mesh::obj_mesh(const char *fName)
{
	FILE *f=fopen(fName,"r");
	char line[1000], stuff[100];
	while (NULL!=fgets(line,999,f)) {
		switch (line[0]) {
		case 'v': { /* Vertex line, like "v 0.1 0.7 0.9" */
			double x,y,z;
			if (4!=sscanf(line,"%99s %lf%lf%lf",stuff,&x,&y,&z)) 
				oglExit("Unparsable v-line '%s'\n",line);
			if (stuff==std::string("v"))
				vtx.push_back(vec3(x,y,z));
			else if (stuff==std::string("vn"))
				norm.push_back(normalize(vec3(x,y,z)));
			else
				oglExit("Unknown v-line '%s'\n",line);
			break;
		}
		case 'f': { /* Face line, like "f 1 2 3 4" or "f 1//1 2//2 3//3 4//4". */
			myFace q;
			for (int i=0;i<4;i++) q.n[i]=0;
			/* FIXME: this is hideous coding, right here... */
			/* Quad face, with normals */
			if (9==sscanf(line,"%99s %d//%d  %d//%d  %d//%d  %d//%d ",stuff,
				&q.v[0],&q.n[0], 
				&q.v[1],&q.n[1], 
				&q.v[2],&q.n[2], 
				&q.v[3],&q.n[3])
			)
				quad.push_back(q);
			/* Quad face, without normals */
			else if (5==sscanf(line,"%99s %d  %d  %d  %d ",stuff,
					&q.v[0], 
					&q.v[1], 
					&q.v[2], 
					&q.v[3])
			)
				quad.push_back(q);
			/* Tri face, with normals */
			else if (7==sscanf(line,"%99s %d//%d  %d//%d  %d//%d ",stuff,
				&q.v[0],&q.n[0], 
				&q.v[1],&q.n[1], 
				&q.v[2],&q.n[2])
			)
				tri.push_back(q);
			/* Tri face, without normals */
			else if (4==sscanf(line,"%99s %d  %d  %d ",stuff,
					&q.v[0], 
					&q.v[1], 
					&q.v[2])
			)
				tri.push_back(q);
			else /* all parse attempts failed! */
				oglExit("Unknown f-line '%s'\n",line);
			break;
		}
		
		default: break; /* ignore unrecognized lines */
		};
	}
}
void obj_mesh::draw_face(const myFace &f,int nVertices)
{
	for (int i=0;i<nVertices;i++)
	{
		if (norm.size()>0) 
			glNormal3fv(norm[f.n[i]-1]);
		glVertex3fv(vtx [f.v[i]-1]);
	}
}
void obj_mesh::draw(void)
{
	glBegin(GL_QUADS);
	for (unsigned int q=0;q<quad.size();q++) draw_face(quad[q],4);
	glEnd();
	glBegin(GL_TRIANGLES);
	for (unsigned int t=0;t<tri.size();t++) draw_face(tri[t],3);
	glEnd();
}
/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down +Z axis; Y points down. */
ortho_frame camera_frame;

/* Location of our camera */
vec3 camera_location=vec3(0,0,1.7);

/* Corresponding key is down */
bool key_is_down[256]={false};

/* Time it took to draw the last frame (for updating physics) */
float seconds_per_frame=0.01;

/* Nice handy timer: returns seconds */
double nice_timer(void) {return glutGet(GLUT_ELAPSED_TIME)*1.0e-3;}

/* Return a random float on [-1,+1] */
float rand_float(void) {
	return (rand()&0xffff)*(1.0/0xffff)*2.0-1.0;
}

/* Return a random two-vector on [-1,+1] */
vec3 rand_vec2(void) {
	return vec3(rand_float(),rand_float(),0.0);
}
/* Return the fractional part of this vector */
vec3 fract(const vec3 &v) {
	return v-vec3(floor(v.x),floor(v.y),floor(v.z));
}

void draw_sphere(const vec3 &pos,float r) {
	glPushMatrix();
	glTranslatef(pos.x,pos.y,pos.z);
	glutSolidSphere(r,12,8);
	glPopMatrix();
}

void my_display(void) {
	glFinish();
	int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	
/* Move the camera */
	float dt=seconds_per_frame;/* time between physics updates, seconds */
	
	float vel=walking_speed; /* velocity, in m/s */
	float forward_dir=0.0, strafe_dir=0.0;
	if (key_is_down['w']) {forward_dir+=1.0;}
	if (key_is_down['s']) {forward_dir-=1.0;}
	if (key_is_down['a']) {strafe_dir-=1.0;}
	if (key_is_down['d']) {strafe_dir+=1.0;}
	
	static vec3 camera_vel = vec3(0.0);
	float friction_frac=1.0;
	vec3 acc = (
		camera_frame.z*forward_dir+
		camera_frame.x*strafe_dir
		)*vel*dt - friction_frac*camera_vel;
	camera_vel = dt*acc + camera_vel;
	
	if (move_camera) 
		camera_location += camera_vel;

	glutPostRedisplay();
	
	/* You are walking: */
	camera_location.z=1.7; /* you can't fly */
	
	camera_frame.x.z=0.0; /* you can't tilt your head */
	camera_frame.y=cross(camera_frame.z,camera_frame.x);
	
	if (false && camera_location.z<1.0) 
	{ /* flight mode: bounce off terrain*/
		camera_location.z=1.0;
		camera_frame.nudge(0,-0.5); /* pull back hard */
	}
	
	camera_frame.orthonormalize();
	
/* Set up camera */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		1000.0 /* Far clipping plane depth */
	);
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(camera_location + camera_frame.z), /* look down +Z */
		VEC3_TO_XYZ(-camera_frame.y) /* OpenGL +Y points up, our Y points down */
	);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	glColor4f(1,1,1,1);
	
/* Set up textures */
	static GLuint monkeyTex, colorTex;
	int texWid=256,texHt=256;
	if (monkeyTex==0) { 
		glGenTextures(1,&monkeyTex); /* make two texture handles */
		glGenTextures(2,&colorTex);
		
		glBindTexture(GL_TEXTURE_2D,monkeyTex);
		readTextureFromFile("monkey.bmp");
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,16);
		
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		
		glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
		
		glActiveTexture(GL_TEXTURE3); /* one texture "unit" */
		glBindTexture(GL_TEXTURE_2D,monkeyTex);
		
		glActiveTexture(GL_TEXTURE0); /* 0'th texture "unit" */
	}
	

	static GLhandleARB prog=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"\n"
	"void main(void) {\n"
	"	reflectance = gl_Color;\n"
	"	worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );\n"
	"	gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; \n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"uniform sampler2D myMonkey, myColors;\n"
	"uniform vec3 camera_vel;\n"
	"\n"
	"void main(void) {\n"
	"	gl_FragColor = 0.125*(\n"
	"		texture2D(myMonkey,vec2(worldCoords))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.125*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.25*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.375*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.5*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.625*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.75*camera_vel))+\n"
	"		texture2D(myMonkey,vec2(worldCoords+0.875*camera_vel))\n"
	"	);\n"
	"	gl_FragColor.a = 1.0; \n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	
	glUniform3fvARB( /* set the "cameraCoords" uniform variable */
		glGetUniformLocationARB(prog, "cameraLoc"),
		1,
		camera_location
	);
	
	glUniform3fvARB(glGetUniformLocationARB(prog,"camera_vel"),1,camera_vel);
	
	glUniform1iARB(glGetUniformLocationARB(prog,"myMonkey"),3);
	glUniform1iARB(glGetUniformLocationARB(prog,"myColors"),7);
	
	/* Draw giant textured plane */
	glPushMatrix();
	glTranslatef(-0.5e3,-0.5e3,0);
	//for (int i=0;i<100;i++) 
	{
		glBegin (GL_TRIANGLE_FAN);
		float z=0.0;
        	glVertex3d(0,0,z); 
        	glVertex3d(1e3,0,z); 
        	glVertex3d(1e3,1e3,z); 
        	glVertex3d(0,1e3,z); 
        	glEnd();
	}
	glPopMatrix();
	
	
	/* Here's a simple diffuse-only shader, for drawing objects */
	static GLhandleARB cprog=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"\n"
	"void main(void) {\n"
	"	reflectance = gl_Color;\n"
	"	worldNormal = gl_NormalMatrix * gl_Normal;\n"
	"	worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );\n"
	"	gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; \n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"uniform vec3 camera_vel;\n"
	"\n"
	"void main(void) {\n"
	"	float d=clamp(dot(worldNormal,vec3(0,0,1)),0.0,1.0);\n"
	"	gl_FragColor = (0.1+0.8*d)*reflectance;\n"
	"	gl_FragColor.a = 1.0; \n"
	"}\n"
	);
	glUseProgramObjectARB(cprog);
	
	glColor3f(1,0,1);
	static obj_mesh potato("potato.obj");
	
	srand(10);
	for (int c=0;c<nPotato;c++) {
		int x=(rand()%1000)-500;
		int y=(rand()%1000)-500;
		if (c==0) { x=0; y=5; /* put first potato in front of user */ }
		glPushMatrix();
		glTranslatef(x,y,1.0); /* float potato off ground */
		glRotatef(last_msec*0.001*1.7,1.0,0.0,0); /* adjust potato axis */
		glRotatef(last_msec*0.001*30,0.0,0.0,1); /* slowly rotate potato */
		potato.draw();
		glPopMatrix();
	}
	
	/* Start the random tree sequence again */
	for (int tree=0;tree<1000;tree++) {
		int x=(rand()%1000)-500;
		int y=(rand()%1000)-500;
		glColor3f(0.3,1.0,0.5);
		glPushMatrix();
		glTranslatef(x,y,0);
		glutSolidCone(1.0, 10.0, 10,10);
		glPopMatrix();
	}
	
	glutPostRedisplay(); /* keeps redisplaying again and again */
	
	glUseProgramObjectARB(0); /* back to fixed-function */


	//glColor3f(0,1,0);
	//glutSolidTorus(0.3,2.0,40,50);

	glBindTexture(GL_TEXTURE_2D,monkeyTex);
	//glCopyTexSubImage2D(GL_TEXTURE_2D,0, 0,0, 0,0, texWid,texHt);

	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	glFinish();
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
	seconds_per_frame=msec_elapsed*1.0e-3;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,speed*dy);
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

void my_keyboard(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going down\n",key);
	key_is_down[key]=true;
	glutPostRedisplay();
}
void my_keyboardup(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going back up\n",key);
	key_is_down[key]=false;
}
int main(int argc,char *argv[]) {
	/* Start camera in reasonable orientation */
	camera_frame.y=vec3(0,0,-1); camera_frame.z=vec3(0,1,0);
	camera_frame.orthonormalize();

	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("Simple Interactive GUI Demo (try the WASD keys!)");
	glewInit();
	
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( win );
	
	new GLUI_Spinner(glui, "Walking Speed: ", &walking_speed);
	new GLUI_Checkbox(glui, "Move camera?", &move_camera);
	new GLUI_Spinner(glui, "Potato Count: ", &nPotato);
	//new GLUI_Column(glui); /* for layout */
	//new GLUI_Rotation(glui,"Light Direction",&light_direction[0][0]);
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	glutKeyboardFunc(my_keyboard);
	glutKeyboardUpFunc(my_keyboardup);
	
	glutMainLoop(); /* processes events */
	return 0;
}
